#!/usr/bin/env python
#
# Copyright (c) 2008  Simon Peter. All rights reserved.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. 
#
# Author(s): Simon Peter <probono@myrealbox.com>

import os
import sys
import gtk
import liveusbgui # a tepache-generated py file
import usbdevice # contains a class which represents USB devices like "/dev/sdb"
import dbus
import subprocess  
import shutil
import atexit
import gettext 
import threading
import urllib2
import getdirsize

from time import sleep
try:
    import gnomeautomount # optional, for temporarily disabling gnome automounting
except:
    pass

os_name = "Ubuntu"
app = "liveusb"
app_name = "Create Live USB system"
app_version = "0.1.1"
app_authors = ["Simon Peter <probono@myrealbox.com>"]
os_versionfile = "/cdrom/.disk/info"
LOCALEDIR = os.path.normpath(os.path.dirname(sys.argv[0]) + "/../locale") # relative to this file
flash_url = "http://fpdownload.macromedia.com/get/flashplayer/current/install_flash_player_10_linux.tar.gz"

def disable_automount():
    try:
        mgr.disable_volume_manager()
    except:
        pass

def enable_automount():
    try:
        mgr.enable_volume_manager()
    except:
        pass

def warning(string):
    print string
    message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING, gtk.BUTTONS_OK, string)
    resp = message.run()
    message.destroy()

def error(string):
    print string
    message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_CANCEL, string)
    resp = message.run()
    message.destroy()

def yesno(string):
    print string
    message = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, gtk.BUTTONS_YES_NO, string)
    response = message.run()
    message.destroy()
    return response

def become_root():
    """Become root and run this script as root. Tested on Ubuntu 8.04 Live
    and Kubuntu 7.10 Live
    """
    if os.getuid() != 0:
        try:
            os.execl("/usr/bin/gksu", "gksu", os.path.abspath(sys.argv[0]))
        except:
            pass
        try:
            os.execl("/usr/bin/kdesu", "-u", "root", "-c",
                     os.path.abspath(sys.argv[0]))
        except: 
            pass
        if os.getuid() != 0:
            error("Could not become root")
            exit(1)

class LiveStickMaker(object):
    """The class that implements the actual USB stick making.
    It should contain no GUI-specific code for reusability.
    Must hence be subclassed by a class that handles 
    all things GUI, like self(progress)...
    """

    def makeLiveStick(self): 
        """Does all the steps necessary to make a bootable live stick"""
        self.checkDependencies()
        result = self.selecteddrive.unmountAllPartitions()
        if result != True:
            error("Could not unmount all partitions on %s" % self.selecteddrive.device)
            return
        self.setStatus("Unmounted all partitions on " + self.selecteddrive.device)
        self.setProgress(0.05)
        if self.checkbutton2.get_active() == True:
            result = self.selecteddrive.makeTwoPartitionsForPersistence()
        else:
            result = self.selecteddrive.makeOneBootableVFAT32Partition()
        if result != True:
            error("Could not partition " + self.selecteddrive.device)
            return
        self.setStatus("Partitioned" + self.selecteddrive.device)
        self.setProgress(0.13)

        # Kubuntu automatically mounts a newly-partitioned disk
        # so we have to unmount it here;
        # better would be to prevent system from automounting 
        # the newly partitioned disk at all (FIXME)
        self.selecteddrive.unmountAllPartitions() 
        result = self.selecteddrive.formatFirstPartitionVFAT32()
        if result != True:
            error("Could not format " + self.selecteddrive.device1)
            return
        self.setStatus("Formatted" + self.selecteddrive.device1)
        if self.checkbutton2.get_active() == True:
            result = self.selecteddrive.formatSecondPartitionExt2()
            if result != True:
                error("Could not format " + self.selecteddrive.device2)
                return
            self.setStatus("Formatted" + self.selecteddrive.device2)            
        self.setProgress(0.1)
        result = self.selecteddrive.writeMBR()
        if result != True:
            error("Could not write Master Boot Record to " + self.selecteddrive.device + "\n" + "(Is syslinux installed?)")
            return
        self.setStatus("Wrote Master Boot Record to " + self.selecteddrive.device)
        self.setProgress(0.15)
        result = self.selecteddrive.mountFirstPartition()
        if result != True:
            error("Could not mount " + self.selecteddrive.device1)
            return 
        tmpmountpoint = self.selecteddrive.mountpoints[0]
        self.tmpmountpoint = tmpmountpoint
        self.setStatus("Mounted " + self.selecteddrive.device1 + " to " + tmpmountpoint)
        self.setProgress(0.2)
        try:
            os.makedirs(tmpmountpoint + "/boot/syslinux")
            subprocess.call(["/bin/sync"])  # otherwise syslinux doesn't see /boot just created
            command = ["syslinux", "-d", "/boot/syslinux", self.selecteddrive.device1]
            subprocess.call(command)
        except:
            error("Could not install syslinux")
            return
        self.setStatus("Installed syslinux on " + self.selecteddrive.device1)
        self.setProgress(0.25)
        r = file("/proc/cmdline","r")
        cmdline = r.read() # by re-using running cmdline, we get the language "for free"
        r.close()
        if cmdline.find("casper") == -1: # use defaults if "casper" is not in our current cmdline
            cmdline = "file=/cdrom/preseed/ubuntu.seed boot=casper " + \
                    "initrd=/casper/initrd.gz quiet splash noprompt -- \n"
        if self.checkbutton2.get_active() == True:
            cmdline = "persistent " + cmdline
        cfg = "DEFAULT /casper/vmlinuz\n" + \
            "APPEND noprompt " + cmdline
        f = file(tmpmountpoint + "/boot/syslinux/syslinux.cfg","w")
        f.write(cfg)
        f.close()
        try:
            self.copy("/cdrom/.disk",tmpmountpoint + "/.disk")
            self.copy("/cdrom/casper",tmpmountpoint + "/casper")
        except:
            error("Could not copy files to " + tmpmountpoint)
            return
        #hideLiveDirectories(self)
        if self.checkbutton1.get_active() == True:
            try:
                self.includeflash()
            except:
                pass
        #enableAllSources(self)
        self.selecteddrive.unmountAllPartitions()
        os.rmdir(tmpmountpoint)
        self.refresh(None) # There are new partitions, so update the combo box
        subprocess.call(["/bin/sync"]) # just to be sure everything is written to disk yet
        self.setStatus("Done, you can remove the device now.")
        self.setProgress(1)
        
    def includeflash(self):
        displayname = "Adobe Flash Player"
        try:
            tmpdir = "/tmp/flashplayer"
            dest = tmpdir + "/flash.tar.gz"
            shutil.rmtree(tmpdir, True)
            os.makedirs(tmpdir)
            self.download(flash_url,dest,displayname)
            saveddir = os.curdir
            os.chdir(tmpdir)
            subprocess.call(["tar", "xfvz", dest])
            os.chdir(saveddir)
            flashdir = self.tmpmountpoint + "/casper/flash.dir/usr/lib/xulrunner-addons/plugins/"
            try:
                os.makedirs(flashdir)
            except:
                pass # could be already there, if our source medium also has it
            shutil.move(tmpdir + "/install_flash_player_10_linux/libflashplayer.so", flashdir + "/libflashplayer.so")
        except:
            error("Could not integrate %s" % displayname)
        shutil.rmtree(tmpdir, True) # clean up
        self.setStatus("Integrated %s" % displayname)
            
    def checkDependencies(self):
        print "Dependency check to be implemented (FIXME)" 

    def checkrunninglive(self):
        """Check if we are running from a live CD (or one is mounted) and give error otherwise"""
        try:
            f = file(os_versionfile,"r")
            self.os_version = f.read()
            f.close()
        except:
            error("Could not open " + os_versionfile + "\nPlease run this application on a " 
                  + os_name + " live system \n"
                  + "or mount a " + os_name + " live ISO to /cdrom")
            exit(1)
        if os.path.exists("/cdrom/casper") == False:
            error("It looks like you are not running this from a live system (/cdrom/casper is missing).\nPlease run this application on a " 
                  + os_name + " live system \n"
                  + "or mount a " + os_name + " live ISO to /cdrom")
            
    def determinecaspersize(self):
        """Determine size of casper/ directory"""
        try:
            result = os.popen('du -sh /cdrom/casper').read()
            self.caspersize = int(str.split(result,"M")[0])
        except:
            error("Could not determine size of /cdrom/casper\nPlease run this application on a " 
                  + os_name + " live system \n"
                  + "or mount a " + os_name + " live ISO to /cdrom")
            exit(1)

class MainDialogSubclass(liveusbgui.Dialog1, LiveStickMaker):
    """A subclass of Dialog1, so that we can add methods here 
    rather than in the tepache-generated file
    """
    def __init__(self):
        self.set_locales()
        self.checkrunninglive()
        self.determinecaspersize()
        liveusbgui.Dialog1.__init__(self)
        self.setStatus(" ") # an empty message makes the status bar the right size
        self.label1.set_text("Create a bootable live USB stick of\n%s"
                             % self.os_version)
        self.dialog1.set_title(app_name)
        self.frame1.set_label("Target device")
        # self.expander1.hide()
        self.expander1.set_label("Options")
        self.checkbutton1.set_label("Download and integrate Adobe Flash Player")
        self.checkbutton2.set_label("Persistent home (allow to save user data)")
        self.checkbutton3.hide()
        self.checkbutton4.hide()
        self.refresh(None)
        
    def set_locales(self):
        # http://www.galileocomputing.de/openbook/python/python_kapitel_15_003.htm
        domain = app
        gettext.bindtextdomain(domain, LOCALEDIR)
        gtk.glade.bindtextdomain(domain, LOCALEDIR)
        gtk.glade.textdomain(domain)
        gettext.textdomain(domain)
        gettext.install(domain, LOCALEDIR)

    def setStatus(self,text):
        """Update the text that is being displayed"""
        print text # to stdout
        self.progressbar1.set_text(text) # to the progress bar
        self.update_gui()

    def pulseProgress(self):
        self.progressbar1.pulse()

    def setProgress(self,value):
        self.progressbar1.set_fraction(value)
        self.update_gui()

    def runcommand(listofargs):
        try:
            retcode = subprocess.call(listofargs)
        except OSError:
            error("Failed to execute " + listofargs[0])
        if retcode != 0:
            error("Failed to execute " + listofargs[0] 
                  + ",\nreturned error code " + str(retcode))
        return retcode

    def enablewidgets(self):
        self.button1.set_sensitive(1)
        self.button2.set_sensitive(1)

    def disablewidgets(self):
        self.button1.set_sensitive(0)
        self.button2.set_sensitive(0)

    def refresh(self,widget):
        """Refresh the combo box of available removable USB devices
        or give error message if no suitable one has been found.
        """
        self.detectRemovableDrives() 
        self.model = self.comboTargets.get_model()
        self.comboTargets.set_model(None)
        self.model.clear()
        for drive in self.drives:
            self.model.append([drive])
        self.comboTargets.set_model(self.model)
        self.comboTargets.set_active(0) # make 1st entry in list selected by default
        if self.drives == []:
            warning("No suitable USB target drive found.\n"
                    + "Please attach the USB target drive and press refresh.")
        self.setProgress(0) # reset status bar when refresh clicked
        self.setStatus(" ") # reset status bar when refresh clicked
        
    def update_gui(self):
        while gtk.events_pending():
            gtk.main_iteration_do(True)

    def start(self,widget):
        """Start button has been pressed, so start doing the main work."""
        self.disablewidgets()
        self.setStatus("Starting to work")
        # would be nicer if we could store arbitrary objects such as USBDevice objects right in the combo box
        # because that does not work, we have to use the index and rely that the order has not changed
        index = self.comboTargets.get_active() # is -1 if there's no active item
        print "index " + str(index)
        if index == -1: # if nothing was selected, index is -1
            warning("No USB target drive selected.\nPlease select a target drive.")
            self.enablewidgets()
            return
        else:
            self.selecteddrive = self.drives[index]
        partitionstring = ""
        safetybuffer = 50 # in MB
        if self.selecteddrive.size < (self.caspersize + safetybuffer): 
            error("This device has less than %i MB capacity.\nIt cannot be used." % (self.caspersize + safetybuffer))
            self.enablewidgets()
            return
        for partition in self.selecteddrive.partitions: 
            partitionstring += " - " + partition + "\n"
        response = yesno("Do you want to install " + os_name + " on\n"  
                         + self.selecteddrive.vendor + " " + self.selecteddrive.product 
                         + " (" + self.selecteddrive.device + ")?\n\n"
                         + "It currently currently contains " 
                         + str(len(self.selecteddrive.partitions)) + " partition(s)\n" 
                         + "which will be deleted.\n" + partitionstring)
        if response != gtk.RESPONSE_YES:
            self.enablewidgets()
            return   
        response = yesno("Are you sure you want to erase\n"  
                         + self.selecteddrive.vendor + " " + self.selecteddrive.product 
                         + " (" + self.selecteddrive.device + ")?")
        if response != gtk.RESPONSE_YES:
            self.enablewidgets()
            return         
        self.setStatus("Starting work on " + str(self.selecteddrive))
        self.update_gui()
        self.makeLiveStick()
        self.enablewidgets()

    def about(self,widget):
        """Show the about box for this application"""
        about = gtk.AboutDialog()
        about.set_authors(app_authors)
        about.set_copyright("Copyright 2008 Simon Peter")
        about.set_wrap_license(0)
        f = file("/usr/share/common-licenses/GPL-2", "r")
        text = f.read()
        f.close()
        about.set_license(text)
        about.set_program_name(app_name)
        about.set_version(app_version)
        about.set_comments("Running on " + self.os_version)
        about.set_logo(gtk.gdk.pixbuf_new_from_file_at_size(
            os.path.dirname(os.path.abspath(sys.argv[0])) + '/../../liveusb/usbpen.png', 
            64, 64))
        about.run()
        about.destroy() # close the about box after it has run

    def detectRemovableDrives(self):
        """Put all removable USB drives that contain no partition 
        that is mounted to /cdrom to self.drives
        """
        # Inspired by code from Luke Macken
        self.drives = [] # empty the list because it gets appended to
        self.bus = dbus.SystemBus()
        hal_obj = self.bus.get_object("org.freedesktop.Hal", "/org/freedesktop/Hal/Manager")
        self.hal = dbus.Interface(hal_obj, "org.freedesktop.Hal.Manager")
        udis = self.hal.FindDeviceByCapability("storage")
        for udi in udis:
            dev = dbus.Interface(self.bus.get_object("org.freedesktop.Hal", udi), "org.freedesktop.Hal.Device")
            if dev.GetProperty("storage.bus") == "usb" and dev.GetProperty("storage.removable"):
                # The properties that we can get via GetProperty are documented on
                # http://webcvs.freedesktop.org/hal/hal/doc/spec/hal-spec.html?view=co#device-properties
                USBDeviceObject = usbdevice.USBDevice(dev.GetProperty("block.device"))
                self.drives.append(USBDeviceObject) 
                # Iterate over mount points on this device 
                # in order to exclude the device that is mounted to /cdrom (which is where we are running from).
                # Need to do this here since unfortunately USBDeviceObject.mountpoints is empty for the 
                # device mounted to /cdrom (smells buggy)
                childudis = self.hal.FindDeviceStringMatch("info.parent",udi)
                for childudi in childudis:
                    child = dbus.Interface(self.bus.get_object("org.freedesktop.Hal", childudi),
                                           "org.freedesktop.Hal.Device")
                    if child.GetProperty("volume.mount_point") == "/cdrom":
                        self.drives.remove(USBDeviceObject)

    def download(self, url, file, displayname):
        down = downloadthread(url, file) 
        down.start()
        # We need to do the actual downloading in a separate thread
        # so that we can freely work on the GUI (progress bar) here.
        # Since down runs in a separate thread, this main app
        # continues with the next steps, so we have to check
        # how the separate thread is doing and delay this main app
        # until the separate thread is done.
        # This also means that error detection and handling must be done in the 
        # other thread - try() in this thread will not catch errors in the
        # other thread.
        self.setProgress(0)
        self.setStatus("Downloading %s..." % displayname)
        progress = down.percent_done()/100
        while down.get_filesize() == 0:
            gtk.main_iteration_do(False)
            sleep(0.01)
            if down.error == True:
                return
        gtk.main_iteration_do(True)
        while progress < 1:
            gtk.main_iteration_do(False)
            sleep(0.01)
            self.setProgress(progress)
            progress = down.percent_done()/100
        gtk.main_iteration_do(True)
        self.setProgress(1)
        self.setStatus("Downloaded %s" % displayname)
        
    def copy(self, source, destination):
        copy = copythread(source, destination)
        copy.start()
        self.pulseProgress() #self.setProgress(0)
        self.setStatus("Copying %s..." % source)
        progress = copy.percent_done()/100
        while copy.get_filesize() == 0:
            gtk.main_iteration_do(False)
            sleep(0.1)
        gtk.main_iteration_do(True)
        siz = getdirsize.getDirSize(source)
        while progress < 1:
            gtk.main_iteration_do(False)
            sleep(2)
            progress = copy.percent_done()/100
            destsiz = getdirsize.getDirSize(destination)
            percdone = float(destsiz)/float(siz)
            self.setProgress(percdone)
        gtk.main_iteration_do(True)
        self.setProgress(1)
        self.setStatus("Please wait, device still active...")
        
class downloadthread(threading.Thread):
    """Give it url and filename and it will download 
    from url and save as filename using a new thread.
    This means that GUI main loop continues to be
    in the driver's seat and proceeds with the next tasks,
    unless it explicitly checks the status e.g., of
    by checking the result of instance.downloaded().
    GUI needs to be handled outside of this thread.
    Based on: http://code.google.com/p/vdown
    """    
    
    def __init__(self, url, file):
        self.file = open(file,'wb')
        self.url = url
        threading.Thread.__init__(self)
        self.done_length = 0
        self.total_length = 100
        self.percent_done = lambda: 100.0*self.done_length/int(self.total_length)
        self.filesize = int(self.total_length)/1024 # in KB
        self.error = False # need this to communicate an error to the other thread
        
    def run(self):
        req = urllib2.Request(self.url)
        # http://www.voidspace.org.uk/python/articles/urllib2.shtml
        req.add_header('User-Agent',app + "/" + app_version)
        try:
            con = urllib2.urlopen(req)
        except: 
            error("Could not download \n%s" % self.url) # return 1
            self.error = True
            exit(1)
        print con.info()
        try:        
            self.total_length = con.info()['Content-length']
        except:
            self.total_length = 0 # Server did not deliver Content-Length
        done_length = 0
        for chunk in con:
            self.done_length += len(chunk)
            self.file.write(chunk)
        self.file.close()
        
    def get_filesize(self):
        return int(self.total_length)/1024

class copythread(threading.Thread):
    """Does copy operation in a new thread, so that we can continue processing
    the GUI in the main app
    """    
    
    def __init__(self, source, destination):
        self.source = source
        self.destination = destination
        threading.Thread.__init__(self)
        self.done_length = 0
        self.total_length = 100
        self.percent_done = lambda: 100.0*self.done_length/int(self.total_length)
        self.filesize = int(self.total_length)/1024 # in KB
        
    def run(self):
        self.total_length = 7777
        shutil.copytree(self.source, self.destination)
        # the 7777 lines (dummy values) need to become replaced with a gradual progress
        # using chunk-wise copying?
        self.done_length = 7777
        
    def get_filesize(self):
        return int(self.total_length)/1024  
    
if __name__ == "__main__":
    try:
        trans = gettext.translation(app)
        trans.install()
    except:
        pass
    become_root()
    mgr = gnomeautomount.gnome_volume_manager()
    disable_automount()
    atexit.register(enable_automount) # ensures this is run at quit time
    MainDialogInstance = MainDialogSubclass()
    MainDialogInstance.run()
